#!/bin/sh
#
# Automate the building of the release
# to do it manually each time.
#
#  This script expects to be executed in the top directory of the 
#  pilot-db project.  It builds one zip files:
#
#
#  pilot-db-<release>-src.tar.gz     - a copy of the current source code
#  pilot-db-<release>.zip  - a copy of all executables,overlays, docs etc
#
#  Usage:   ./build-release <version> <configure flags>
#
#  e.g:   ./build-snapshot 1.1
#
#
#



# Get the release number (e.g. '1.1' or '2.0')

if [ "x$1" = "x" ]; then
	echo "Usage: $0 <release-number>"
	exit
else
	release="$1"
fi

#  get rid of the first argument... now $@ refers to what is unprocessed
shift 1

die() {
	echo $1
	exit
}



#
# Make a release directory structure based on the name
#
release_dir="pilot-db-${release}"
[ ! -e $release_dir ] || die "The release directory ${release_dir} exists."
plugin_release_dir="pilot-db-${release}-plugins"
[ ! -e $plugin_release_dir ] || die "The plugin release directory ${plugin_release_dir} exists."


# Form the filename that we will use.
ver=`echo $1 | sed -e 's/\./_/g'`
prczip="../pilot-db-${release}.zip"
pluginzip="../pilot-db-${release}-plugins.zip"
srczip="pilot-db-${release}-src.tar.gz"


[ ! -f $prczip ] || die "The destination file ${prczip} already exists."
[ ! -f $pluginzip ] || die "The destination file ${pluginzip} already exists."
[ ! -f ../$srczip ] || die "The destination file ../${srczip} already exists."

mkdir $release_dir
mkdir ${release_dir}/overlay

mkdir ${plugin_release_dir}



#
#  potential versions of the db application
#
light="--enable-light"
mid="--enable-mid"
standard="--enable-standard"


#
#  Here is a list of the builds that go into the release
#
builds="light mid standard"

#
#  Files and directories to be copied into the distribution
#
files="NEWS ChangeLog README* docs"



# Do some initial cleanups.
if [ -f Makefile ]; then
	make distclean 
	[ $? == 0 ] || die "couldn't clean the distribution"
fi

# temporary files
hfile=${release_dir}/__HEAD__
bfile=${release_dir}/__BIX__



# build various versions of db
for build in ${builds}; do

	if [ -f Makefile ]; then
		make clean
		[ $? == 0 ] || die "couldn't clean the build"
	fi
	echo "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++"
	echo "+ Making ${build}"
	echo "+ Options: ${!build} $@"
	echo "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++"

	# Pass along any extra arguments sent to this script...
	./configure ${!build} $@
	make 
	[ $? == 0 ] || die " **  Build Failed!  (db-${build}) ** "

	mv db*.prc ${release_dir}/db-${build}.prc
	builtfiles="$builtfiles db-${build}.prc"
	date=`date`
	if [ -f ${hfile} ]; then
		echo "ABORTING, the file ${hfile} must be moved out of the way"
		exit
	fi
	if [ -f ${bfile} ]; then
		echo "ABORTING, the file ${bfile} must be moved out of the way"
		exit
	fi


	# record the configuration options in BuildInfo file along with the prcs
	echo -e "\n\n---     =================     ---\ndb-${build} was built on $date with:\n" >> ${hfile}

	if [ -f ${release_dir}/BuildInfo ]; then
        	cat ${release_dir}/BuildInfo ${hfile} src/configuration > ${bfile}
        	mv ${bfile} ${release_dir}/BuildInfo
	else
		cat ${hfile} src/configuration > ${release_dir}/BuildInfo
   	fi
	rm ${hfile}
done


make uninstall.prc
[ $? == 0 ] || die " **  Build Failed!  (uninstall) ** "
mv uninstall.prc ${release_dir}/
[ $? == 0 ] || die " **  Couldn't move (or find) uninstall.prc! ** "

make misc 
[ $? == 0 ] || die "  ** Build Failed!  (misc .prcs) **"
mv *.prc ${release_dir}/overlay/

make all-plugins
[ $? == 0 ] || die "  ** Build Failed!  (plugins) **"
find plugins -name "*.prc" | xargs -n1 -iqq mv qq ${plugin_release_dir}

for f in ${files}; do
	cp -r ${f} ${release_dir}
done

# Create the zip file with all the executables
zip -r ${prczip} ${release_dir}
[ $? == 0 ] || die "couldn't make zip file."
rm -rf ${release_dir}

# Create the zip file with all the executables
zip -r ${pluginzip} ${plugin_release_dir}
[ $? == 0 ] || die "couldn't make plugin zip file."
rm -rf ${plugin_release_dir}


# Clean up
make distclean

# Create the source zip file
dir=`pwd | sed -e "s/.*\/\([^\/]*\)$/\1/"`
cd ..
echo "${srczip} ... ${dir}"
tar -cvz ${dir} -f ${srczip}
[ $? == 0 ] || die "couldn't make source zip file."











